//=========================================================
// CaeX_YEPStatusMenuExt.js
//=========================================================

/*:
 * @plugindesc v1.1 - Lets you add more attributes to Yanfly's Status Menu Core.
 * @author Caethyril
 *
 * @help Plugin Commands:
 *   None.
 *
 * Help:
 *   Define new attributes to display by adding to the Custom Attributes list.
 *    - Key is the string used to determine where it shows in YEP_StatusMenuCore
 *    - Label is the name of the attribute, shown to the player in-game
 *    - Value is a formula determining the value of the attribute.
 *   Then just add the new key(s) via the plugin parameters of YEP_StatusMenuCore.
 *
 *   Example formulae (for gold & item drop rates):
 *    - var nme = new Game_Enemy(1,0,0); nme.dropItemRate()
 *        This will return the item drop rate for enemy #1.
 *    - var trp = new Game_Troop(); trp.setup(1); trp.goldRate()
 *        This will return the gold drop rate for troop #1.
 *   Note that typically these rates are identical for all enemies/troops.
 *
 * Compatibility:
 *   Must be loaded after Yanfly's Status Menu Core plugin.
 *   Aliases Yanfly's drawAttributeData method on Window_StatusInfo.
 *
 * Terms of use:
 *   Free to use and modify.
 *
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * Update log:
 *   1.1: Redesigned, now allows arbitrary new attributes to be added! \o/
 *        Users must design their own formulae for the values, though.
 *        Added a couple of formula examples to the help description.
 *   1.0: Initial release.
 *
 * @param Custom Attributes
 * @text Custom Attributes
 * @type struct<CaeTypeCustomAttr>[]
 * @desc List of custom values to add to the Attributes page of Yanfly's Status Menu Core.
 * @default []
 */
// ==================================================
/*~struct~CaeTypeCustomAttr:
 * @param Key
 * @text Key
 * @type text
 * @desc The code for use in Yanfly's Status Menu Core plugin parameters.
 * @default
 *
 * @param Label
 * @text Label
 * @type text
 * @desc How this entry will be labelled in-game.
 * @default
 *
 * @param Value
 * @text Value
 * @type text
 * @desc Formula determining the value to display.
 * Attributes will always display as percentages (1 = 100%).
 * @default
 */

var Imported = Imported || {};				// Import namespace, var can redefine
Imported.CaeX_YEPStatusMenuExt = 1.1;			// Import declaration

var CAE = CAE || {};					// Author namespace, var can redefine
CAE.X_YEPStatusMenuExt = CAE.X_YEPStatusMenuExt || {};	// Plugin namespace

(function(_) {

'use strict';

	if (!Imported.YEP_StatusMenuCore) throw new Error('CaeX_YEPStatusMenuExt.js error! Could not find required plugin YEP_StatusMenuCore.js.');

	// Convert string -> array of {Key, Label, Value} objects
	_.params = JSON.parse(PluginManager.parameters('CaeX_YEPStatusMenuExt')['Custom Attributes']).map(JSON.parse);

	_.Window_StatusInfo_drawAttributeData = Window_StatusInfo.prototype.drawAttributeData;	// Alias
	Window_StatusInfo.prototype.drawAttributeData = function(attr, dx, dy, dw) {
		let ix = _.params.map(function(o) { return o.Key; }).indexOf(attr);		// Match key from plugin params
		if (ix < 0) {									// If no match
			_.Window_StatusInfo_drawAttributeData.call(this, attr, dx, dy, dw);	// Call original routine
		} else {									// Else
			let label = _.params[ix].Label;						// Get param label
			let valEv = _.params[ix].Value;						// Get param value string
			let value = 0;								// Initialise for scope
			try {
				value = eval(valEv);						// Attempt eval
			} catch (ex) {								// On error, log message to console
				console.error('CaeX_YEPStatusMenuExt.js error! Attribute eval failed:\n' + valEv);
				value = 0;							// Then zero value
			}
			this.drawAttributeName(label, dx, dy, dw);				// Draw label
			this.drawAttributeRate(value, dx, dy, dw);				// Draw value
		}
	};

})(CAE.X_YEPStatusMenuExt);